import numpy as np
import json
from sklearn.metrics import f1_score, mean_squared_error
from sklearn.model_selection import cross_val_score

# lets try a QML approach based on a random feature map
from squlearn.encoding_circuit import LayeredEncodingCircuit
from squlearn.kernel.matrix import FidelityKernel, ProjectedQuantumKernel
from squlearn.util import Executor
from squlearn.kernel import QSVC, QGPR, QSVR
from sklearn.svm import SVC, SVR
from sklearn.kernel_ridge import KernelRidge

# potential squlearn qnn import!
import random

import os

# Environment for reinforcement training
import gym
from gym import spaces
from scipy.linalg import sqrtm

import matplotlib.pyplot as plt

from .abstract_game import AbstractGame


# define action space

actions = [  # action integer number
    "",  # 2
    "H",  # 3
    r"Rx(x;=np.pi*(x),{x})",  # 4
    r"Ry(x;=np.pi*(x),{x})",  # 5
    r"Rz(x;=np.pi*(x),{x})",  # 6
    "cx",  # 13
    "cy",
    "cz",
    r"Rx(x;=np.arctan(x),{x})",  # 14
    "X",  # 15
    "Y",  # 16
    "Z",  # 17
    r"Rx(p;=0*p+3/4*np.pi,{p})",  # 18
    r"Ry(p;=0*p+3/4*np.pi,{p})",  # 19
    r"Rz(p;=0*p+3/4*np.pi,{p})",  # 20
    r"crx(p;=0*p+3/4*np.pi,{p})",  # 21
    r"cry(p;=0*p+3/4*np.pi,{p})",  # 22
    r"crz(p;=0*p+3/4*np.pi,{p})",  # 23
    r"Rx(p;=0*p+1/2*np.pi,{p})",  # 24
    r"Ry(p;=0*p+1/2*np.pi,{p})",  # 25
    r"Rz(p;=0*p+1/2*np.pi,{p})",  # 26
    r"crx(p;=0*p+1/2*np.pi,{p})",  # 27
    r"cry(p;=0*p+1/2*np.pi,{p})",  # 28
    r"crz(p;=0*p+1/2*np.pi,{p})",  # 29
    r"Rx(p;=0*p+1/4*np.pi,{p})",  # 30
    r"Ry(p;=0*p+1/4*np.pi,{p})",  # 31
    r"Rz(p;=0*p+1/4*np.pi,{p})",  # 32
    r"crx(p;=0*p+1/4*np.pi,{p})",  # 33
    r"cry(p;=0*p+1/4*np.pi,{p})",  # 34
    r"crz(p;=0*p+1/4*np.pi,{p})",  # 35
    r"Rx(p;=0*p+1/8*np.pi,{p})",  # 36
    r"Ry(p;=0*p+1/8*np.pi,{p})",  # 37
    r"Rz(p;=0*p+1/8*np.pi,{p})",  # 38
    r"crx(p;=0*p+1/8*np.pi,{p})",  # 39
    r"cry(p;=0*p+1/8*np.pi,{p})",  # 40
    r"crz(p;=0*p+1/8*np.pi,{p})"  # 41
    r"Rx(p;=0*p+5/4*np.pi,{p})",  # 36
    r"Ry(p;=0*p+5/4*np.pi,{p})",  # 37
    r"Rz(p;=0*p+5/4*np.pi,{p})",  # 38
    r"crx(p;=0*p+5/4*np.pi,{p})",  # 39
    r"cry(p;=0*p+5/4*np.pi,{p})",  # 40
    r"crz(p;=0*p+5/4*np.pi,{p})"  # 41
    r"Rx(p;=0*p+1.5*np.pi,{p})",  # 36
    r"Ry(p;=0*p+1.5*np.pi,{p})",  # 37
    r"Rz(p;=0*p+1.5*np.pi,{p})",  # 38
    r"crx(p;=0*p+1.5*np.pi,{p})",  # 39
    r"cry(p;=0*p+1.5*np.pi,{p})",  # 40
    r"crz(p;=0*p+1.5*np.pi,{p})",  # 41
]

# ToDo: include gamma parameter of the PQK
# possible initial guess for gamma: 1 / max_diff_feature_space_rbf_kernel**2

num_actions = len(actions)


# set up a dummy Env with some default parameters
# write wrapper function which instantly returns Game class
class ScoreEnv(gym.Env):
    def __init__(
        self,
        qml_model,
        scoring,
        training_data,
        training_labels,
        num_qubits,
        num_features,
        max_depth,
        storage,
        random_start=0.2,
        test_data=None,
        test_labels=None,
        pure_test=False,
    ):
        """
        Initialization function of the environment
        """

        super(ScoreEnv, self).__init__()
        # Define action and observation space (integer arrays)

        # Defines the possible actions
        self.actions = actions
        self.action_space = spaces.Discrete(len(self.actions))

        # Observations feature map converted to integer string
        self.observation_space = spaces.Box(
            low=0, high=len(self.actions) - 1, shape=(max_depth,), dtype=int
        )

        self._qml_model = qml_model
        self._storage_manager = storage
        self._max_depth = max_depth
        self._random_start = random_start
        self._scoring = scoring

        self.training_data = training_data
        self.training_labels = training_labels
        self.test_data = test_data
        self.test_labels = test_labels
        self.pure_test = pure_test

        self.best_cv_score = np.inf
        self.best_fm_str = ""
        self.tol = 1e-5
        self.seed = np.random.randint(0, 1000)
        # Dictionary for string to integer labels
        self.action_dic = {}
        i = 0
        for a in self.actions:
            self.action_dic[a] = i
            i += 1

        # try to calculate the best possible classical kernel matrix in the init -> then it just has to be
        # evaluated once as long as the training data stays the same
        self._num_qubits = num_qubits
        self._num_features = num_features


        files = os.listdir(os.path.abspath("frames"))

        # Iterate through each file
        for file in files:
            if file.endswith(".png"):
                file_path = os.path.join("frames", file)
                os.remove(file_path)

    def reset(self):
        """
        Reset function of the environment
        Resets the feature map to initial state
        """
        self.fm_str = ""
        self.last_action = ""
        self.done = False
        self.steps_done = 0
        self.last_cv_score = np.inf

        # Randomly select an option from restart_architectures with a 10% chance
        if random.random() < self._random_start:
            num_gates = self._max_depth // 2
            possible_actions = [action for action in self.actions if "(x)" in action]
            chosen_action = random.choice(possible_actions)
            for _ in range(num_gates - 1):
                action = random.choice(self.actions)
                self.fm_str += action + "-"
            self.fm_str += chosen_action + "-"
            # print("random fm restart:", self.fm_str)

        self.reward = 0.0
        self.observations = self.text_to_int(self.fm_str)

        return self.observations

    def step(self, action):
        """
        Function that performs an action and returns the reward and the resulting
        feature-map as observation space
        """
        self.steps_done += 1

        # pred_error = np.inf
        # var_kernel = np.inf
        cv_score = 0.0
        punish_x = 0.0
        punish_action = 0.0
        reward_cv_score = 0.0

        # Add gates to the feature map
        self.fm_str = self.fm_str + self.actions[action] + "-"

        # print('step with respective feature map:', self.steps_done, self.fm_str)
        if self.steps_done >= 1 and action == self.last_action:
            # print("same action as before!",self.fm_str)
            punish_action = -1.0

        if "(x)" not in self.fm_str:
            # Capture the case, that there is no X in the circuit
            punish_x = -1.0
            # print("No X in the circuit!",self.fm_str)

        elif self.pure_test:
            punish_x = 10.0
            # print("self.fm_str",self.fm_str)

        else:

            lfm = LayeredEncodingCircuit.from_string(
                encoding_circuit_str=self.fm_str,
                num_qubits=self._num_qubits,
                num_features=self._num_features,
            )
            quantum_kernel = ProjectedQuantumKernel(
                encoding_circuit=lfm,
                executor=Executor("statevector_simulator"),
                parameter_seed=0,
                gamma=0.5,
            )

            if self._qml_model == "qsvc":
                # qml_model = QSVC(quantum_kernel=quantum_kernel)
                qml_model = SVC(kernel=quantum_kernel.evaluate)
            elif self._qml_model == "qsvr":
                # qml_model = QSVR(quantum_kernel=quantum_kernel)
                qml_model = SVR(kernel=quantum_kernel.evaluate, C=10, epsilon=0.01)

            # either worst case or np.mean
            # use WORST CASE
            # changed from min to mean!!
            cv_score = cross_val_score(
                estimator=qml_model,
                scoring=self._scoring,
                X=self.training_data,
                y=self.training_labels,
                cv=5,
            )
            cv_score = np.mean(cv_score)

            # cv_score = np.mean(cv_score)
            cv_score = -1.0 * cv_score
            if self._scoring == "f1" and cv_score == 0.0:
                reward_cv_score = -50.0
                # print('zero or negative f1score!')
                self.done = True
            elif cv_score <= (self.last_cv_score - self.tol):
                # print('improved cv score!:', self._scoring, cv_score, self.last_cv_score, self.fm_str, self.steps_done)
                reward_cv_score = 50.0
                if cv_score <= (self.best_cv_score - self.tol):
                    self.best_cv_score = cv_score
                    self.best_fm_str = self.fm_str
                    # print('better overall cv score!:', self._scoring, cv_score, self.best_cv_score, self.best_fm_str, self.steps_done)
                    reward_cv_score += 100.0

            else:
                punish_action = -1.0

            self.last_cv_score = cv_score

        # probably the split of the reward in that way makes no sense because both of them dont happen at the same time anyway
        self.reward = reward_cv_score + punish_x + punish_action

        self.last_action = action
        # Create observation integer array
        self.observations = self.text_to_int(self.fm_str)

        # both if conditions follow self._max_depth since for our problem max steps = max depth
        # If too many steps are done, finish this environment
        if self.steps_done >= self._max_depth:
            self.done = True
        # If the feature map is to long, finish this environment
        if np.count_nonzero(self.observations) >= self._max_depth:
            self.done = True

        if self.done == True and self.best_cv_score != np.inf and self.best_fm_str != "":
            self._storage_manager.update_storage(
                {"best_fm_str": [self.best_fm_str], "cv_score": [self.best_cv_score]}
            )
            # self._storage_manager.update_storage(self.best_fm_str, self.best_pred_error)
            # print('updated storage dict at the end of the steps!', self.best_fm_str, self.best_cv_score)
            self._storage_manager.save_storage()

        info = {}
        # Return featuremap as observation, current reward for action, done (=true if environment is done)
        return self.observations, self.reward, self.done, info

    def text_to_int(self, text: str):
        """
        Helper function for generating the observation integer array from the feature_map string
        """
        text_array = text.split("-")
        text_array.pop()
        observation = np.zeros(self._max_depth, dtype=int)

        i = 0
        for t in text_array:
            if t != "":
                observation[i] = self.action_dic[t]
                i = i + 1
        return observation

    def legal_actions(self):
        """
        Should return the legal actions at each turn, if it is not available, it can return
        the whole action space. At each turn, the game have to be able to handle one of returned actions.

        Returns:
            An array of integers, subset of the action space.
        """
        return list(range(len(self.actions)))


class Game(AbstractGame):
    """
    Game wrapper.
    """

    def __init__(
        self, seed=None
    ):  # actions, num_qubits, num_features, max_num_gates, X_train, y_train, storage, seed=None):

        # self.storage_manager = None
        self.env = ScoreEnv

        self.frame_counter = 0
        self.run_counter = 0
        self.action_counter = 0



    def step(self, action):
        """
        Apply action to the game.

        Args:
            action : action of the action_space to take.

        Returns:
            The new observation, the reward and a boolean if the game has ended.
        """
        observation, reward, done, _ = self.env.step(action)
        return np.array([[observation]]), reward, done

    def legal_actions(self):
        """
        Should return the legal actions at each turn, if it is not available, it can return
        the whole action space. At each turn, the game have to be able to handle one of returned actions.

        For complex game where calculating legal moves is too long, the idea is to define the legal actions
        equal to the action space but to return a negative reward if the action is illegal.

        Returns:
            An array of integers, subset of the action space.
        """

        return self.env.legal_actions()

    def reset(self):
        """
        Reset the game for a new game.

        Returns:
            Initial observation of the game.
        """
        # self.storage_manager = self.env._storage_manager
        self.run_counter += 1
        self.action_counter = 0
        return np.array([[self.env.reset()]])

    def close(self):
        """
        Properly close the game.
        """

        pass
        # print("closed")

    def to_play(self):
        """
        Return the current player.

        Returns:
            The current player, it should be an element of the players list in the config.
        """
        return 0

    def render(self):
        """
        Display the game observation.
        """

        lfm = LayeredEncodingCircuit.from_string(
            encoding_circuit_str=self.env.fm_str,
            num_qubits=self.env._num_qubits,
            num_features=self.env._num_features,
        )

        filename = f"frames/frame_{self.frame_counter:03d}.png"
        filename = os.path.abspath(filename)

        plt.figure()
        # plt.autoscale(tight=True)
        plot = lfm.draw("mpl", style="clifford", fold=-1)

        title_str1 = f"Run: {self.run_counter}"
        title_str2 = f"Step: {self.action_counter}"

        plot.axes[0].text(0.2, 1.0, title_str1, fontsize=20, ha="left", va="center")
        plot.axes[0].text(4.0, 1.0, title_str2, fontsize=20, ha="left", va="center")

        plot.savefig(filename, dpi=100, bbox_inches="tight", pad_inches=0.0)

        plt.close()

        self.frame_counter += 1
        self.action_counter += 1

    def import_storage(file_name):
        import json

        with open(file_name, "r") as json_file:
            loaded_data = json.load(json_file)
        best_pred_error_list = loaded_data["cv_score"]
        best_circuit = loaded_data["best_fm_str"]
        # Find the index of the minimum value in the list
        min_index = min(enumerate(best_pred_error_list), key=lambda x: x[1])[0]
        # Get the minimum value and best circuit
        min_value = best_pred_error_list[min_index]
        best_circuit = best_circuit[min_index]
        # print("best cv score:", min_value)
        # print("best circuit:", best_circuit)

        return loaded_data


def create_game(env, seed=None, render_mode="file", canvas=None):
    class GameWrapper(Game):
        def __init__(self, seed=None):
            super().__init__(seed)
            self.env = env

    return GameWrapper


class StorageManager:
    def __init__(self, result_dict, filename):
        self.storage_dict = result_dict
        self._filename = filename

    def update_storage(self, new_data):
        for key in new_data:
            if key in self.storage_dict:
                # Check if new_data[key] is already in storage_dict[key]
                if new_data[key] not in self.storage_dict[key]:
                    self.storage_dict[key].extend(new_data[key])
                    # Ensure any numpy arrays are converted to lists before serializing to JSON
                    self.storage_dict[key] = [
                        entry.tolist() if isinstance(entry, np.ndarray) else entry
                        for entry in self.storage_dict[key]
                    ]

    def save_storage(self):
        # Check for unnecessary double entries
        for key in self.storage_dict:
            self.storage_dict[key] = list(set(self.storage_dict[key]))

        with open(self._filename, "w") as json_file:
            json.dump(self.storage_dict, json_file)

    def load_storage(self, file_name):
        with open(file_name, "r") as json_file:
            loaded_data = json.load(json_file)
        for key in loaded_data:
            if key in self.storage_dict:
                self.storage_dict[key] = loaded_data[key]
